/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <DbObject.h>

/*---------------------------------------------------------------------------*/
static int VectorCount         = 0;
static int VectorContentsCount = 0;
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
DbVector* DbVectorNew (DbType element)
/*---------------------------------------------------------------------------*/
{
  DbVector* object = NULL;
  int element_size;

  if (element == NULL) return (NULL);

  element_size = DbTypeSize (element);
  if (element_size == 0) return (NULL);

  object = (DbVector*) malloc (sizeof (DbVector));
  VectorCount++;

  DbVectorInit (object, element);

  return (object);
}

/*---------------------------------------------------------------------------*/
void DbVectorDelete (DbVector* object)
/*---------------------------------------------------------------------------*/
{
  DbVectorErase (object);

  free (object);
  VectorCount--;
}

/*---------------------------------------------------------------------------*/
void DbVectorInit (DbVector* object, DbType element)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return;

  object->element_handle = NULL;
  object->size           = 0;
  object->contents       = NULL;

  object->element_handle = DbTypeHandleNew ();
  DbHandleSet (object->element_handle, element);
}

/*---------------------------------------------------------------------------*/
void DbVectorErase (DbVector* object)
/*---------------------------------------------------------------------------*/
{
  DbType element;

  element = DbVectorGetElementType (object);
  if (element != NULL)
    {
      int n;
      char* address = (char*) object->contents;
      int element_size = DbTypeSize (element);

      for (n = 0; n < object->size; n++)
        {
          DbTypeErase (element, address);
          address += element_size;
        }
    }

  if (object->contents != NULL)
    {
      free (object->contents);
      VectorContentsCount--;
      object->contents = NULL;
    }

  object->size = 0;

  DbTypeHandleRemove (object->element_handle);
  object->element_handle = NULL;
}

/*---------------------------------------------------------------------------*/
void DbVectorClear (DbVector* object)
/*---------------------------------------------------------------------------*/
{
  DbType element;

  element = DbVectorGetElementType (object);
  if (element != NULL)
    {
      int n;
      char* address = (char*) object->contents;
      int element_size = DbTypeSize (element);

      for (n = 0; n < object->size; n++)
        {
          DbTypeErase (element, address);
          address += element_size;
        }
    }

  if (object->contents != NULL)
    {
      free (object->contents);
      VectorContentsCount--;
      object->contents = NULL;
    }

  object->size = 0;
}

/*---------------------------------------------------------------------------*/
DbType DbVectorGetElementType (DbVector* object)
/*---------------------------------------------------------------------------*/
{
  if (object == NULL) return (NULL);
  if (object->element_handle == NULL) return (NULL);

  return ((DbType) DbHandleGet (object->element_handle));
}

/*---------------------------------------------------------------------------*/
void DbVectorExtend (DbVector* object, int size)
/*---------------------------------------------------------------------------*/
{
  char* begin;
  char* end;
  int bytes;
  DbType element;

  element = DbVectorGetElementType (object);
  if (element == NULL) return;

  if (size > object->size)
    {
      if (object->contents == NULL)
        {
          bytes = size * DbTypeSize (element);
          object->contents = (void*) malloc (bytes);
          VectorContentsCount++;
          begin = (char*) object->contents;
        }
      else
        {
          int old_size = object->size * DbTypeSize (element);

          bytes = size * DbTypeSize (element);

          object->contents = (void*) realloc (object->contents, bytes);

          begin = (char*) object->contents;
          begin += old_size;
          bytes -= old_size;
        }
      memset (begin, 0, bytes);

      end = ((char*) object->contents) + (size * DbTypeSize (element));
      
      while (begin < end)
        {
          DbTypeInit (element, begin);
          begin += DbTypeSize (element);
        }
      
      object->size = size;
    }
}

/*---------------------------------------------------------------------------*/
int DbVectorGetSize (DbVector* object)
/*---------------------------------------------------------------------------*/
{
  DbType element;

  element = DbVectorGetElementType (object);
  if (element == NULL) return (0);

  return (object->size);
}

/*---------------------------------------------------------------------------*/
int DbVectorAddElement (DbVector* object)
/*---------------------------------------------------------------------------*/
{
  int size;

  size = DbVectorGetSize (object);
  size++;
  DbVectorExtend (object, size);

  return (size);
}

/*---------------------------------------------------------------------------*/
void DbVectorRemoveElement (DbVector* object, int index)
/*---------------------------------------------------------------------------*/
{
  char* address;
  int element_size;
  DbType element;

  if (object == NULL) return;

  if ((index < 0) || (index >= object->size)) return;

  address = (char*) object->contents;
  if (address == NULL) return;

  element = DbVectorGetElementType (object);
  element_size = DbTypeSize (element);

  if (element_size == 0) return;

  address += index * element_size;
  DbTypeErase (element, address);

  for (; index < object->size - 1; index++)
    {
      memcpy (address, address + element_size, element_size);
      address += element_size;
    }

  object->size--;

  if (object->size == 0)
    {
      free (object->contents);
      VectorContentsCount--;
      object->contents = NULL;
    }
  else
    {
      object->contents = (void*) realloc (object->contents, 
                                          object->size * element_size);
    }
}

/*---------------------------------------------------------------------------*/
void* DbVectorGetAddress (DbVector* object, int index)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType element;

  element = DbVectorGetElementType (object);
  if (element == NULL) return (NULL);

  if (object->contents == NULL) return (NULL);
  if ((index < 0) ||
      (index >= object->size)) return (NULL);

  result = (char*) object->contents;

  return (result + (index * DbTypeSize (element)));
}

