#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>
#include <CmMessage.h>

/*---------------------------------------------------------------------*/
typedef struct
{
  DbError error;
  char error_text[256];
  void* object;
} TransactionInfo;
/*---------------------------------------------------------------------*/

CmMessageStatus DBVersionHandler (CmMessage message);

/*---------------------------------------------------------------------*/
DbObject DbParserCreateObject (char* text);
static char** Split (const char* text);
static char* shell_execute (const char* command);

/*--------------------------------------------------------*/
static int Done = 0;

/*--------------------------------------------------------*/
static void help ()
/*--------------------------------------------------------*/
{
  fprintf (stderr, " > db command arg [ command arg ] ... \n");
  fprintf (stderr, "\n");
  fprintf (stderr, "     create_object   <object definition>  "
           ": create an object\n");
  fprintf (stderr, "     create_object -f <definition file>   "
           ": create an object from a file\n");
  fprintf (stderr, "     load_object     <object name>        "
           ": load an object\n");
  fprintf (stderr, "     load_objects    <pattern>            "
           ": load some objects\n");
  fprintf (stderr, "     change_object   <new definition>     "
           ": change the value of an object\n");
  fprintf (stderr, "     change_object -f <file name>         "
           ": change the value of an object from a file\n");
  fprintf (stderr, "     delete_object   <object name>        "
           ": delete an object\n");
  fprintf (stderr, "     create_type     <ddl file>           "
           ": create a type\n");
  fprintf (stderr, "     load_type       <type name>          "
           ": load a type\n");
  fprintf (stderr, "     delete_type     <type name>          "
           ": delete a type\n");
  fprintf (stderr, "     object_names    <pattern>            "
           ": query object names\n");
  fprintf (stderr, "     path_names      <pattern>            "
           ": query path names\n");
  fprintf (stderr, "     update_header   <type name>          "
           ": update header for a type\n");
  fprintf (stderr, "     update_headers                       "
           ": update all headers\n");
  fprintf (stderr, "     stop_server                          "
           ": stop the DbServer\n");
  fprintf (stderr, "     restart_server                       "
           ": restart the DbServer\n");
  fprintf (stderr, "     domains                              "
           ": Show all domains\n");
  fprintf (stderr, "     domain [<domain>]                    "
           ": Show domain infos\n");
  fprintf (stderr, "     set_domain [-csh|-sh] <domain>       "
           ": Set domain infos\n");
  fprintf (stderr, "     version                              "
           ": Show Db version\n");
  fprintf (stderr, "     wait            <seconds>            "
           ": insert a wait between 2 commands\n");
  fprintf (stderr, "");
}

/*---------------------------------------------------------------------------*/
static char* build_full_name (char* object_name, char* path_name)
/*---------------------------------------------------------------------------*/
{
  static char full_name[512];
  
  if ((path_name != NULL) &&
      (strlen (path_name) > 0))
    {
      strcpy (full_name, path_name);
      strcat (full_name, "/");
    }
  else
    {
      full_name[0] = 0;
    }
  
  strcat (full_name, object_name);
  
  return (full_name);
}

/*---------------------------------------------------------------------*/
static void load_object (char* name)
/*---------------------------------------------------------------------*/
{
  DbObject object;
  char* definition;
  
  object = DbLoadObject (name);
  definition = DbConvertObjectToString (object);
  printf ("%s\n", definition);
}

/*---------------------------------------------------------------------*/
static void load_objects (char* pattern)
/*---------------------------------------------------------------------*/
{
  static CLList list = NULL;
  CIterator it;
  DbObject object;
  char* definition;

  if (list == NULL)
    {
      GCReferenceAssign (list, CLListNew ());
    }
  else
    {
      CLListErase (list);
    }

  DbLoadObjects (pattern, list);
  it = CLListGiveIterator (list);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      object = (DbObject) CIteratorItem (it);
      definition = DbConvertObjectToString (object);
      printf ("%s\n", definition);
    }
  CIteratorDelete (it);
}

/*---------------------------------------------------------------------*/
static void change_object (char* definition)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  DbObject object = NULL;
  char* name;
  char* path_name;

  for (;;)
    {
      DbClearErrorInfo ();
      
      GCReferenceAssign (object, DbParserCreateObject (definition));
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;

          GCReferenceClear (object);
        }
      else
        {
          break;
        }
    }

  if (DbGetError () != DbErrorOk)
    {
      GCReferenceClear (object);
      DbPrintError ();
      return;
    }

  name = DbObjectGetName (object);
  path_name = DbObjectGetPathName (object);
  name = build_full_name (name, path_name);

  GCReferenceClear (object);

  DbSetLocks (DbLockReadWrite);

  object = DbLoadObject (name);
  if (object != NULL)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      transaction = CmOpenTransaction ("DbChangeObject", &info);

      message = CmMessageNew ();
  
      CmMessagePutText (message, definition);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbCommitObject");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbChangeObject>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
    }
  else
    {
      DbPrintError ();
    }

  DbResetLocks ();
}

/*---------------------------------------------------------------------*/
static void create_type (char* definition)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  message = CmMessageNew ();
  CmMessagePutText (message, definition);

  CmMessageSetType (message, "DbCreateType");

  transaction = CmOpenTransaction ("DbCreateType", &info);

  CmMessagePutInt (message, transaction);

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "Cannot contact the DbServer");
    }
  else
    {
      DbWaitForTransaction (transaction);
    }
  
  CmCloseTransaction (transaction);
  CmMessageDelete (message);
  DbSetErrorInfo (info.error, info.error_text);
  
  DbPrintError ();
}

/*---------------------------------------------------------------------*/
static void delete_type (char* name)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  message = CmMessageNew ();
  CmMessagePutText (message, name);
  CmMessageSetType (message, "DbDeleteType");

  transaction = CmOpenTransaction ("DbDeleteType", &info);

  CmMessagePutInt (message, transaction);

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "Cannot contact the DbServer");
    }
  else
    {
      DbWaitForTransaction (transaction);
    }
  
  CmCloseTransaction (transaction);
  CmMessageDelete (message);
  DbSetErrorInfo (info.error, info.error_text);
  
  DbPrintError ();
}

/*---------------------------------------------------------------------*/
static void update_header (char* name)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  message = CmMessageNew ();
  CmMessagePutText (message, name);
  CmMessageSetType (message, "DbUpdateType");

  transaction = CmOpenTransaction ("DbUpdateType", &info);

  CmMessagePutInt (message, transaction);

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "Cannot contact the DbServer");
    }
  else
    {
      DbWaitForTransaction (transaction);
    }
  
  CmCloseTransaction (transaction);
  CmMessageDelete (message);
  DbSetErrorInfo (info.error, info.error_text);
  
  DbPrintError ();
}

/*---------------------------------------------------------------------*/
static void update_headers ()
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  message = CmMessageNew ();
  CmMessageSetType (message, "DbUpdateHeaders");

  transaction = CmOpenTransaction ("DbUpdateHeaders", &info);

  CmMessagePutInt (message, transaction);

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "Cannot contact the DbServer");
    }
  else
    {
      DbWaitForTransaction (transaction);
    }
  
  CmCloseTransaction (transaction);
  CmMessageDelete (message);
  DbSetErrorInfo (info.error, info.error_text);
  
  DbPrintError ();
}

/*--------------------------------------------------------*/
int main (int argc, char* argv[])
/*--------------------------------------------------------*/
{
  int has_answer = 0;
  CmMessage message;
  char* destination = NULL;

  if (!CmMessageOpenMultipleServer ("db"))
    {
      fprintf (stderr, "Cannot connect to Cm\n");
      exit (0);
    }

  DbStartup ();

  argc--;
  argv++;

  destination = DbServerName ();
  message = CmMessageNew ();

  if (argc == 0)
    {
      help ();
    }

  while (argc > 0)
    {
      char arg[80];

      strcpy (arg, argv[0]);

      if (!strcmp (arg, "create_object"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char definition[1024];

	      if (!strcmp (argv[0], "-f"))
		{
		  argc--;
		  argv++;
		  
		  if (argc > 0)
		    {
		      char* text;

		      strcpy (definition, argv[0]);

		      text = DbReadFile (definition);
		      if (text != NULL)
			{
			  DbCreateObject (text);
			  free (text);
			}
		    }
		}
	      else
		{
		  strcpy (definition, argv[0]);

		  DbCreateObject (definition);
		}
	    }
        }
      else if (!strcmp (arg, "load_object"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char name[256];

              strcpy (name, argv[0]);

              load_object (name);
            }
        }
      else if (!strcmp (arg, "load_objects"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char pattern[256];

              strcpy (pattern, argv[0]);

              load_objects (pattern);
            }
        }
      else if (!strcmp (arg, "change_object"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char definition[1024];

	      if (!strcmp (argv[0], "-f"))
		{
		  argc--;
		  argv++;
		  
		  if (argc > 0)
		    {
		      char* text;

		      strcpy (definition, argv[0]);

		      text = DbReadFile (definition);
		      if (text != NULL)
			{
			  printf ("new def = [%s]\n", text);

			  change_object (text);

			  free (text);
			}
		    }
		}
	      else
		{
		  strcpy (definition, argv[0]);

		  printf ("new def = [%s]\n", definition);

		  change_object (definition);
		}
            }
        }
      else if (!strcmp (arg, "delete_object"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char name[256];

              strcpy (name, argv[0]);

              DbDeleteObject (name);
              DbPrintError ();
            }
        }
      else if (!strcmp (arg, "create_type"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char name[256];
              char* text;

              strcpy (name, argv[0]);
              text = DbReadFile (name);

              if (text != NULL)
                {
                  create_type (text);
                  free (text);
                }
              else
                {
                  fprintf (stderr, "Cannot open file %s\n", name);
                }

            }
        }
      else if (!strcmp (arg, "delete_type"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char name[256];

              strcpy (name, argv[0]);

              delete_type (name);
            }
        }
      else if (!strcmp (arg, "load_type"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char name[256];
	      char* text;

              sprintf (name, "%s.ddl", argv[0]);
              text = DbReadFile (name);

              if (text != NULL)
                {
		  printf ("%s\n", text);
                  free (text);
                }
              else
                {
                  fprintf (stderr, "Cannot open file %s\n", name);
                }
            }
        }
      else if (!strcmp (arg, "object_names"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char expression[256];
              char* names;
              char** words;

              strcpy (expression, argv[0]);

              names = DbObjectGetNames (expression);
              words = Split (names);
              if (words != NULL)
                {
                  int i = 0;

                  while (words[i] != NULL)
                    {
                      printf ("%s\n", words[i]);
                      i++;
                    }
                  free (words);
                }
            }
        }
      else if (!strcmp (arg, "path_names"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char expression[256];
              char* names;

              strcpy (expression, argv[0]);

              names = DbSubPathGetNames (expression);
              DbPrintError ();
              if (names != NULL) printf ("%s\n", names);
            }
        }
      else if (!strcmp (arg, "update_header"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              char name[256];

              strcpy (name, argv[0]);

              update_header (name);
            }
        }
      else if (!strcmp (arg, "update_headers"))
        {
	  destination = NULL;
          update_headers ();
        }
      else if (!strcmp (arg, "stop_server"))
        {
          char* domain;
          char* server_name;
          char answer[80];
          char* login;

	  destination = NULL;

          domain = getenv ("CMDOMAIN");
          server_name = DbServerName ();
	  login = shell_execute ("whoami");

	  if (strcmp (login, "virgod") && !strcmp (domain, "Cascina"))
	    {
	      printf ("User %s is not authorized to stop server %s in domain Cascina\n", 
		      login, server_name);
	      CmMessageDelete (message);
	      return (0);
	    }
	  else
	    {
	      if (!strcmp (domain, "Cascina"))
		{
		  
		  printf ("  #    #    ##    #####   #    #     #    #    #   ####  \n");
		  printf ("  #    #   #  #   #    #  ##   #     #    ##   #  #    # \n");
		  printf ("  #    #  #    #  #    #  # #  #     #    # #  #  #      \n");
		  printf ("  # ## #  ######  #####   #  # #     #    #  # #  #  ### \n");
		  printf ("  ##  ##  #    #  #   #   #   ##     #    #   ##  #    # \n");
		  printf ("  #    #  #    #  #    #  #    #     #    #    #   ####  \n\n");
		  
		  printf ("   ####     ##     ####    ####      #    #    #    ##   \n");
		  printf ("  #    #   #  #   #       #    #     #    ##   #   #  #  \n");
		  printf ("  #       #    #   ####   #          #    # #  #  #    # \n");
		  printf ("  #       ######       #  #          #    #  # #  ###### \n");
		  printf ("  #    #  #    #  #    #  #    #     #    #   ##  #    # \n");
		  printf ("   ####   #    #   ####    ####      #    #    #  #    # \n");

		}

	      printf ("You are going to stop the server %s of the domain %s\n",
		      server_name, domain);
	      printf ("Do you really want to do this ? y/[n] ");
	      fflush (stdout);
	      answer[0] = 0;
	      gets (answer);
	      
	      if ((answer[0] == 'y') ||
		  (answer[0] == 'Y'))
		{
		  DbStopServer ();
		}
	    }
        }
      else if (!strcmp (arg, "restart_server"))
        {
	  char* domain;
          char* server_name;
          char answer[80];

	  destination = NULL;

          domain = getenv ("CMDOMAIN");

          server_name = DbServerName ();

	  printf ("You are going to restart the server %s of the domain %s\n",
                  server_name, domain);
          printf ("Do you really want to do this ? y/[n] ");

          fflush (stdout);
          answer[0] = 0;
          gets (answer);

          if ((answer[0] == 'y') ||
              (answer[0] == 'Y'))
            {
	      DbRestartServer ();
            }

        }
      else if (!strcmp (arg, "domains"))
        {
          char domains_file[256];
          char* db_root;
          FILE* domains;
          char line[256];
          char all_domains[256] = "";
          int quiet = 0;

	  destination = NULL;

          argc--;
          argv++;

          while (argc > 0)
            {
              char* arg = argv[0];

              if (!strcmp (arg, "-quiet"))
                {
                  quiet = 1;
                }

              argc--;
              argv++;
            }

          db_root = getenv ("DBROOT");
          sprintf (domains_file,"%s/mgr/DbDomains", db_root);
              
          domains = fopen (domains_file, "r");
          if (domains != NULL)
            {
              while (fgets (line, sizeof(line), domains) != NULL)
                {
                  char** words;
                  char* nl;
                  
                  nl = strchr (line, '\n');
                  if (nl != NULL) *nl = 0;
                  
                  words = Split (line);
                  if (words != NULL)
                    {
                      if (words[0][0] != '#')
                        {
                          strcat (all_domains, words[0]);
                          strcat (all_domains, "\n");
                        }

                      free (words);
                    }
                }
              
              fclose (domains);
            }
              
          if (!quiet) printf ("Possible Db domains are\n");
          printf ("%s\n", all_domains);
        }
      else if (!strcmp (arg, "domain"))
        {
          char domain[256] = "";
          char db_server_name[256] = "";
          char db_cm_domain[256] = "";
          char db_objects_root[256] = "";
          char db_cvs_root[256] = "";
          char domains_file[256];
          char* db_root;
          FILE* domains;
          char line[256];
          char all_domains[256] = "";
          int quiet = 0;

	  destination = NULL;

          argc--;
          argv++;

          while (argc > 0)
            {
              char* arg = argv[0];

              if (!strcmp (arg, "-quiet"))
                {
                  quiet = 1;
                }
              else
                {
                  strcpy (domain, argv[0]);
                }
              argc--;
              argv++;
            }

          if (strlen (domain) == 0)
            {
              char* env;

              env = getenv ("DBDOMAIN");
              if (env != NULL)
                {
                  strcpy (domain, env);
                }
            }

          db_root = getenv ("DBROOT");
          sprintf (domains_file,"%s/mgr/DbDomains", db_root);
              
          domains = fopen (domains_file, "r");
          if (domains != NULL)
            {
              while (fgets (line, sizeof(line), domains) != NULL)
                {
                  char** words;
                  char* nl;
                  
                  nl = strchr (line, '\n');
                  if (nl != NULL) *nl = 0;
                  
                  words = Split (line);
                  if (words != NULL)
                    {
                      if (!strcmp (words[0], domain))
                        {
                          strcpy (db_cm_domain, words[1]);
                          strcpy (db_server_name, words[2]);
                          strcpy (db_objects_root, words[3]);
                          strcpy (db_cvs_root, words[4]);
                          break;
                        }
                      
                      if (words[0][0] != '#')
                        {
                          strcat (all_domains, words[0]);
                          strcat (all_domains, "\n");
                        }

                      free (words);
                    }
                }
              
              fclose (domains);
            }
              
          if (strlen (domain) == 0)
            {
              printf ("The Db domain is NOT defined.\n");
              printf ("Please define it using the dbdomain command among\n"
                      "%s\n",
                      all_domains);
            }
          else
            {
              char* server_name;
              char* cm_domain;
              char* objects_root;
              char* cvs_root;
              CmConnect server = NULL;
              char* host;

              cm_domain = getenv ("CMDOMAIN");
              objects_root = getenv ("DBOBJECTSROOT");
              cvs_root = getenv ("DBCVSROOT");
              server_name = DbServerName ();

              if (!strcmp (cm_domain, db_cm_domain))
                {
                  server = CmConnectNew (db_server_name);
                }

              printf ("Definition of the %s domain\n", domain);
              printf (" Cm domain     : %s\n", db_cm_domain);
              printf (" DbServer      : %s", db_server_name);

              if (!strcmp (cm_domain, db_cm_domain))
                {
                  if (server == NULL)
                    {
                      printf (" (not running)\n");
                    }
                  else
                    {
                      printf (" (running on %s)\n", CmConnectGetHost (server));
                    }
                }
              else
                {
                  printf (" (not in current Cm domain)\n");
                }
              
              printf (" DBOBJECTSROOT : %s\n", db_objects_root);
              printf (" DBCVSROOT     : %s\n", db_cvs_root);
            }
        }
      else if (!strcmp (arg, "set_domain"))
        {
          char db_server_name[256] = "";
          char db_cm_domain[256] = "";
          char db_objects_root[256] = "";
          char db_cvs_root[256] = "";
          char domain[256] = "";
          char domains_file[256];
          char* db_root;
          FILE* domains;
          char line[256];
          char all_domains[256] = "";
          enum { csh, sh } mode = csh;

	  destination = NULL;

          argc--;
          argv++;

          while (argc > 0)
            {
              if (!strcmp (argv[0], "-csh"))
                {
                  mode = csh;
                }
              else if (!strcmp (argv[0], "-sh"))
                {
                  mode = sh;
                }
              else
                {
                  strcpy (domain, argv[0]);
                }
              argc--;
              argv++;
            }

          if (strlen (domain) == 0)
            {
              char* env;

              env = getenv ("DBDOMAIN");

              if (env != NULL)
                {
                  strcpy (domain, env);
                }
            }

          db_root = getenv ("DBROOT");
          sprintf (domains_file,"%s/mgr/DbDomains", db_root);
          
          domains = fopen (domains_file, "r");
          if (domains != NULL)
            {
              while (fgets (line, sizeof(line), domains) != NULL)
                {
                  char** words;
                  char* nl;
                  
                  nl = strchr (line, '\n');
                  if (nl != NULL) *nl = 0;
                  
                  words = Split (line);
                  if (words != NULL)
                    {
                      if (!strcmp (words[0], domain))
                        {
                          strcpy (db_cm_domain, words[1]);
                          strcpy (db_server_name, words[2]);
                          strcpy (db_objects_root, words[3]);
                          strcpy (db_cvs_root, words[4]);
                          break;
                        }
                      
                      if (words[0][0] != '#')
                        {
                          strcat (all_domains, words[0]);
                          strcat (all_domains, "\n");
                        }
                      
                      free (words);
                    }
                }
              
              fclose (domains);
            }
          
          if (strlen (domain) == 0)
            {
              fprintf (stderr, "The Db domain is NOT defined.\n");
              fprintf (stderr, "Please select one among %s.\n",
                       all_domains);
            }
          else
            {
              if (strlen (db_server_name) > 0)
                {
                  switch (mode)
                    {
                      case csh :
                        printf ("setenv DBDOMAIN %s;"
                                " setenv CMDOMAIN %s;"
                                " setenv DBSERVERNAME %s;"
                                " setenv DBOBJECTSROOT %s;"
                                " setenv DBCVSROOT %s\n", 
                                domain,
                                db_cm_domain,
                                db_server_name,
                                db_objects_root,
                                db_cvs_root);
                        break;
                      case sh :
                        printf ("DBDOMAIN=%s; export DBDOMAIN;"
                                " CMDOMAIN=%s; export CMDOMAIN;"
                                " DBSERVERNAME=%s; export DBSERVERNAME;"
                                " DBOBJECTSROOT=%s; export DBOBJECTSROOT;"
                                " DBCVSROOT=%s; export DBCVSROOT\n", 
                                domain,
                                db_cm_domain,
                                db_server_name,
                                db_objects_root,
                                db_cvs_root);
                        break;
                    }
                }
              else
                {
                  fprintf (stderr, "The domain %s is not defined."
                           " Please select one among\n%s",
                           domain,
                           all_domains);
                }
            }
        }
      else if (!strcmp (arg, "version"))
        {
	  /*          char* version;

		      argc--;
		      argv++;
		      
		      version = shell_execute ("basename $DBROOT");
		      
		      printf ("%s\n", version);

	  */
	  has_answer = 1;
	  CmMessageInstallHandler ((CmMessageHandler) DBVersionHandler,
				   "DbVersion");
	  CmMessageSetType (message, "DbGetVersion");
	}

      else if (!strcmp (arg, "wait"))
        {
	  destination = NULL;
          argc--;
          argv++;

          if (argc > 0)
            {
              double timeout;

              strcpy (arg, argv[0]);

              sscanf (arg, "%lg", &timeout);

              CmMessageWaitWithTimeout (timeout);
            }
        }
      else
        {
	  destination = NULL;
          help ();
          break;
        }
      
      if (DbGetError () != DbErrorOk)
        {
          DbPrintError ();
        }

      argc--;
      argv++;
    }

  if (destination != NULL)
    {
      CmMessageSend (message, destination);

      if (has_answer)
        {
          while (!Done)
            {
              CmMessageWait ();
            }
        }
      else
        {
          CmMessageWaitWithTimeout (0.001);
        }
    }
  
  CmMessageDelete (message);
  return (0);
}

/*--------------------------------------------------------------*/
CmMessageStatus DBVersionHandler (CmMessage message)
/*--------------------------------------------------------------*/
{
  char* name;

  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      name = CmMessageGetText (message);
      printf ("Version: %s\n", name);
    }

  Done = 1;

  return (CmMessageBreak);
}

/*----------------------------------------------------------*/
static char** Split (const char* text)
/*----------------------------------------------------------*/
/*  Split a line into words. Separators are spaces and tabs */
/*  Text enclosed in double quotes is one word.             */
/*----------------------------------------------------------*/
{
  char** result = NULL;
  int texts = 0;
  int finished = 0;
  char* t;

  if (text == NULL) return (NULL);

  t = (char*) malloc (strlen (text) + 1);
  strcpy (t, text);

  while (strlen (text) != 0)
    {
      char* space;

      while ((*t == ' ') ||
             (*t == '\t')) t++;

      if ((t[0] == '"') ||
          (t[0] == '\'') ||
          (t[0] == ':'))
        {
          char sep = t[0];
          char* quote;

          t++;
          quote = strchr (t, sep);
          if (quote != NULL) *quote = 0;
          else finished = 1;
        }
      else
        {
          int offset;

          offset = strcspn (t, " \t");
          if (offset < 0 || t[offset] == 0) finished = 1;
          if (!finished)
            {
              space = t + offset;
              *space = 0;
            }
        }

      if (result == NULL)
        {
          texts = 1;
          result = (char**) malloc ((texts + 1) * sizeof (char*));
        }
      else
        {
          texts++;
          result = (char**) realloc (result, 
                                     (texts + 1) * sizeof (char*));
        }

      result[texts - 1] = t;
      result[texts] = NULL;

      if (finished) break;

      t += strlen (t) + 1;
    }

  return (result);
}

/*----------------------------------------------------------*/
static char* shell_execute (const char* command)
/*----------------------------------------------------------*/
{
  static char output[256];
  FILE* file;

  output[0] = 0;

  file = popen (command, "r");
  if (file == NULL) return (output);

  if (fgets (output, sizeof (output), file) != NULL)
    {
      char* nl;
      nl = strchr (output, '\n');
      if (nl != NULL) *nl = 0;
    }

  pclose (file);

  return (output);
}
